#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest import mock
from unittest.mock import call

from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.network.v2 import subnet as subnet_v2
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes_v3
from openstackclient.tests.unit.network.v2 import fakes as network_fakes
from openstackclient.tests.unit import utils as tests_utils


class TestSubnet(network_fakes.TestNetworkV2):

    def setUp(self):
        super(TestSubnet, self).setUp()

        # Get a shortcut to the network client
        self.network = self.app.client_manager.network
        # Get a shortcut to the ProjectManager Mock
        self.projects_mock = self.app.client_manager.identity.projects
        # Get a shortcut to the DomainManager Mock
        self.domains_mock = self.app.client_manager.identity.domains


class TestCreateSubnet(TestSubnet):

    def _init_subnet_variables(self):
        self.project = identity_fakes_v3.FakeProject.create_one_project()
        self.domain = identity_fakes_v3.FakeDomain.create_one_domain()
        # An IPv4 subnet to be created with mostly default values
        self._subnet = network_fakes.FakeSubnet.create_one_subnet(
            attrs={
                'project_id': self.project.id,
            }
        )

        # Subnet pool to be used to create a subnet from a pool
        self._subnet_pool = \
            network_fakes.FakeSubnetPool.create_one_subnet_pool()

        # An IPv4 subnet to be created using a specific subnet pool
        self._subnet_from_pool = network_fakes.FakeSubnet.create_one_subnet(
            attrs={
                'project_id': self.project.id,
                'subnetpool_id': self._subnet_pool.id,
                'dns_nameservers': ['8.8.8.8',
                                    '8.8.4.4'],
                'host_routes': [{'destination': '10.20.20.0/24',
                                 'nexthop': '10.20.20.1'},
                                {'destination': '10.30.30.0/24',
                                 'nexthop': '10.30.30.1'}],
                'service_types': ['network:router_gateway',
                                  'network:floatingip_agent_gateway'],
            }
        )

        # An IPv6 subnet to be created with most options specified
        self._subnet_ipv6 = network_fakes.FakeSubnet.create_one_subnet(
            attrs={
                'project_id': self.project.id,
                'cidr': 'fe80:0:0:a00a::/64',
                'enable_dhcp': True,
                'dns_nameservers': ['fe80:27ff:a00a:f00f::ffff',
                                    'fe80:37ff:a00a:f00f::ffff'],
                'allocation_pools': [{'start': 'fe80::a00a:0:c0de:0:100',
                                      'end': 'fe80::a00a:0:c0de:0:f000'},
                                     {'start': 'fe80::a00a:0:c0de:1:100',
                                      'end': 'fe80::a00a:0:c0de:1:f000'}],
                'host_routes': [{'destination': 'fe80:27ff:a00a:f00f::/64',
                                 'nexthop': 'fe80:27ff:a00a:f00f::1'},
                                {'destination': 'fe80:37ff:a00a:f00f::/64',
                                 'nexthop': 'fe80:37ff:a00a:f00f::1'}],
                'ip_version': 6,
                'gateway_ip': 'fe80::a00a:0:c0de:0:1',
                'ipv6_address_mode': 'slaac',
                'ipv6_ra_mode': 'slaac',
                'subnetpool_id': 'None',
                'service_types': ['network:router_gateway',
                                  'network:floatingip_agent_gateway'],
            }
        )

        # The network to be returned from find_network
        self._network = network_fakes.create_one_network(
            attrs={
                'id': self._subnet.network_id,
            }
        )

        # The network segment to be returned from find_segment
        self._network_segment = \
            network_fakes.create_one_network_segment(
                attrs={
                    'network_id': self._subnet.network_id,
                }
            )

        self.columns = (
            'allocation_pools',
            'cidr',
            'description',
            'dns_nameservers',
            'enable_dhcp',
            'gateway_ip',
            'host_routes',
            'id',
            'ip_version',
            'ipv6_address_mode',
            'ipv6_ra_mode',
            'name',
            'network_id',
            'project_id',
            'segment_id',
            'service_types',
            'subnetpool_id',
            'tags',
        )

        self.data = (
            subnet_v2.AllocationPoolsColumn(self._subnet.allocation_pools),
            self._subnet.cidr,
            self._subnet.description,
            format_columns.ListColumn(self._subnet.dns_nameservers),
            self._subnet.enable_dhcp,
            self._subnet.gateway_ip,
            subnet_v2.HostRoutesColumn(self._subnet.host_routes),
            self._subnet.id,
            self._subnet.ip_version,
            self._subnet.ipv6_address_mode,
            self._subnet.ipv6_ra_mode,
            self._subnet.name,
            self._subnet.network_id,
            self._subnet.project_id,
            self._subnet.segment_id,
            format_columns.ListColumn(self._subnet.service_types),
            self._subnet.subnetpool_id,
            format_columns.ListColumn(self._subnet.tags),
        )

        self.data_subnet_pool = (
            subnet_v2.AllocationPoolsColumn(
                self._subnet_from_pool.allocation_pools),
            self._subnet_from_pool.cidr,
            self._subnet_from_pool.description,
            format_columns.ListColumn(self._subnet_from_pool.dns_nameservers),
            self._subnet_from_pool.enable_dhcp,
            self._subnet_from_pool.gateway_ip,
            subnet_v2.HostRoutesColumn(self._subnet_from_pool.host_routes),
            self._subnet_from_pool.id,
            self._subnet_from_pool.ip_version,
            self._subnet_from_pool.ipv6_address_mode,
            self._subnet_from_pool.ipv6_ra_mode,
            self._subnet_from_pool.name,
            self._subnet_from_pool.network_id,
            self._subnet_from_pool.project_id,
            self._subnet_from_pool.segment_id,
            format_columns.ListColumn(self._subnet_from_pool.service_types),
            self._subnet_from_pool.subnetpool_id,
            format_columns.ListColumn(self._subnet_from_pool.tags),
        )

        self.data_ipv6 = (
            subnet_v2.AllocationPoolsColumn(
                self._subnet_ipv6.allocation_pools),
            self._subnet_ipv6.cidr,
            self._subnet_ipv6.description,
            format_columns.ListColumn(self._subnet_ipv6.dns_nameservers),
            self._subnet_ipv6.enable_dhcp,
            self._subnet_ipv6.gateway_ip,
            subnet_v2.HostRoutesColumn(self._subnet_ipv6.host_routes),
            self._subnet_ipv6.id,
            self._subnet_ipv6.ip_version,
            self._subnet_ipv6.ipv6_address_mode,
            self._subnet_ipv6.ipv6_ra_mode,
            self._subnet_ipv6.name,
            self._subnet_ipv6.network_id,
            self._subnet_ipv6.project_id,
            self._subnet_ipv6.segment_id,
            format_columns.ListColumn(self._subnet_ipv6.service_types),
            self._subnet_ipv6.subnetpool_id,
            format_columns.ListColumn(self._subnet_ipv6.tags),
        )

    def setUp(self):
        self._init_subnet_variables()
        super(TestCreateSubnet, self).setUp()

        # Get the command object to test
        self.cmd = subnet_v2.CreateSubnet(self.app, self.namespace)

        self.projects_mock.get.return_value = self.project
        self.domains_mock.get.return_value = self.domain

        # Mock SDK calls for all tests.
        self.network.create_subnet = mock.Mock(return_value=self._subnet)
        self.network.set_tags = mock.Mock(return_value=None)
        self.network.find_network = mock.Mock(return_value=self._network)
        self.network.find_segment = mock.Mock(
            return_value=self._network_segment
        )
        self.network.find_subnet_pool = mock.Mock(
            return_value=self._subnet_pool
        )

    def test_create_no_options(self):
        arglist = []
        verifylist = []

        # Testing that a call without the required argument will fail and
        # throw a "ParserExecption"
        self.assertRaises(tests_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)
        self.assertFalse(self.network.create_subnet.called)
        self.assertFalse(self.network.set_tags.called)

    def test_create_default_options(self):
        # Mock SDK calls for this test.
        self._network.id = self._subnet.network_id

        arglist = [
            "--subnet-range", self._subnet.cidr,
            "--network", self._subnet.network_id,
            self._subnet.name,
        ]
        verifylist = [
            ('name', self._subnet.name),
            ('subnet_range', self._subnet.cidr),
            ('network', self._subnet.network_id),
            ('ip_version', self._subnet.ip_version),
            ('gateway', 'auto'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.create_subnet.assert_called_once_with(**{
            'cidr': self._subnet.cidr,
            'ip_version': self._subnet.ip_version,
            'name': self._subnet.name,
            'network_id': self._subnet.network_id,
        })
        self.assertFalse(self.network.set_tags.called)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_from_subnet_pool_options(self):
        # Mock SDK calls for this test.
        self.network.create_subnet.return_value = self._subnet_from_pool
        self.network.set_tags = mock.Mock(return_value=None)
        self._network.id = self._subnet_from_pool.network_id

        arglist = [
            self._subnet_from_pool.name,
            "--subnet-pool", self._subnet_from_pool.subnetpool_id,
            "--prefix-length", '24',
            "--network", self._subnet_from_pool.network_id,
            "--ip-version", str(self._subnet_from_pool.ip_version),
            "--gateway", self._subnet_from_pool.gateway_ip,
            "--dhcp",
        ]

        for dns_addr in self._subnet_from_pool.dns_nameservers:
            arglist.append('--dns-nameserver')
            arglist.append(dns_addr)

        for host_route in self._subnet_from_pool.host_routes:
            arglist.append('--host-route')
            value = 'gateway=' + host_route.get('nexthop', '') + \
                    ',destination=' + host_route.get('destination', '')
            arglist.append(value)

        for service_type in self._subnet_from_pool.service_types:
            arglist.append('--service-type')
            arglist.append(service_type)

        verifylist = [
            ('name', self._subnet_from_pool.name),
            ('prefix_length', '24'),
            ('network', self._subnet_from_pool.network_id),
            ('ip_version', self._subnet_from_pool.ip_version),
            ('gateway', self._subnet_from_pool.gateway_ip),
            ('dns_nameservers', self._subnet_from_pool.dns_nameservers),
            ('dhcp', self._subnet_from_pool.enable_dhcp),
            ('host_routes', subnet_v2.convert_entries_to_gateway(
                self._subnet_from_pool.host_routes)),
            ('subnet_pool', self._subnet_from_pool.subnetpool_id),
            ('service_types', self._subnet_from_pool.service_types),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.create_subnet.assert_called_once_with(**{
            'dns_nameservers': self._subnet_from_pool.dns_nameservers,
            'enable_dhcp': self._subnet_from_pool.enable_dhcp,
            'gateway_ip': self._subnet_from_pool.gateway_ip,
            'host_routes': self._subnet_from_pool.host_routes,
            'ip_version': self._subnet_from_pool.ip_version,
            'name': self._subnet_from_pool.name,
            'network_id': self._subnet_from_pool.network_id,
            'prefixlen': '24',
            'subnetpool_id': self._subnet_from_pool.subnetpool_id,
            'service_types': self._subnet_from_pool.service_types,
        })
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data_subnet_pool, data)

    def test_create_options_subnet_range_ipv6(self):
        # Mock SDK calls for this test.
        self.network.create_subnet.return_value = self._subnet_ipv6
        self._network.id = self._subnet_ipv6.network_id

        arglist = [
            self._subnet_ipv6.name,
            "--subnet-range", self._subnet_ipv6.cidr,
            "--network", self._subnet_ipv6.network_id,
            "--ip-version", str(self._subnet_ipv6.ip_version),
            "--ipv6-ra-mode", self._subnet_ipv6.ipv6_ra_mode,
            "--ipv6-address-mode", self._subnet_ipv6.ipv6_address_mode,
            "--gateway", self._subnet_ipv6.gateway_ip,
            "--dhcp",
        ]

        for dns_addr in self._subnet_ipv6.dns_nameservers:
            arglist.append('--dns-nameserver')
            arglist.append(dns_addr)

        for host_route in self._subnet_ipv6.host_routes:
            arglist.append('--host-route')
            value = 'gateway=' + host_route.get('nexthop', '') + \
                    ',destination=' + host_route.get('destination', '')
            arglist.append(value)

        for pool in self._subnet_ipv6.allocation_pools:
            arglist.append('--allocation-pool')
            value = 'start=' + pool.get('start', '') + \
                    ',end=' + pool.get('end', '')
            arglist.append(value)

        for service_type in self._subnet_ipv6.service_types:
            arglist.append('--service-type')
            arglist.append(service_type)

        verifylist = [
            ('name', self._subnet_ipv6.name),
            ('subnet_range', self._subnet_ipv6.cidr),
            ('network', self._subnet_ipv6.network_id),
            ('ip_version', self._subnet_ipv6.ip_version),
            ('ipv6_ra_mode', self._subnet_ipv6.ipv6_ra_mode),
            ('ipv6_address_mode', self._subnet_ipv6.ipv6_address_mode),
            ('gateway', self._subnet_ipv6.gateway_ip),
            ('dns_nameservers', self._subnet_ipv6.dns_nameservers),
            ('dhcp', self._subnet_ipv6.enable_dhcp),
            ('host_routes', subnet_v2.convert_entries_to_gateway(
                self._subnet_ipv6.host_routes)),
            ('allocation_pools', self._subnet_ipv6.allocation_pools),
            ('service_types', self._subnet_ipv6.service_types),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.create_subnet.assert_called_once_with(**{
            'cidr': self._subnet_ipv6.cidr,
            'dns_nameservers': self._subnet_ipv6.dns_nameservers,
            'enable_dhcp': self._subnet_ipv6.enable_dhcp,
            'gateway_ip': self._subnet_ipv6.gateway_ip,
            'host_routes': self._subnet_ipv6.host_routes,
            'ip_version': self._subnet_ipv6.ip_version,
            'ipv6_address_mode': self._subnet_ipv6.ipv6_address_mode,
            'ipv6_ra_mode': self._subnet_ipv6.ipv6_ra_mode,
            'name': self._subnet_ipv6.name,
            'network_id': self._subnet_ipv6.network_id,
            'allocation_pools': self._subnet_ipv6.allocation_pools,
            'service_types': self._subnet_ipv6.service_types,
        })
        self.assertFalse(self.network.set_tags.called)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data_ipv6, data)

    def test_create_with_network_segment(self):
        # Mock SDK calls for this test.
        self._network.id = self._subnet.network_id

        arglist = [
            "--subnet-range", self._subnet.cidr,
            "--network-segment", self._network_segment.id,
            "--network", self._subnet.network_id,
            self._subnet.name,
        ]
        verifylist = [
            ('name', self._subnet.name),
            ('subnet_range', self._subnet.cidr),
            ('network_segment', self._network_segment.id),
            ('network', self._subnet.network_id),
            ('ip_version', self._subnet.ip_version),
            ('gateway', 'auto'),

        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.create_subnet.assert_called_once_with(**{
            'cidr': self._subnet.cidr,
            'ip_version': self._subnet.ip_version,
            'name': self._subnet.name,
            'network_id': self._subnet.network_id,
            'segment_id': self._network_segment.id,
        })
        self.assertFalse(self.network.set_tags.called)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_description(self):
        # Mock SDK calls for this test.
        self._network.id = self._subnet.network_id

        arglist = [
            "--subnet-range", self._subnet.cidr,
            "--network", self._subnet.network_id,
            "--description", self._subnet.description,
            self._subnet.name,
        ]
        verifylist = [
            ('name', self._subnet.name),
            ('description', self._subnet.description),
            ('subnet_range', self._subnet.cidr),
            ('network', self._subnet.network_id),
            ('ip_version', self._subnet.ip_version),
            ('gateway', 'auto'),

        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.create_subnet.assert_called_once_with(**{
            'cidr': self._subnet.cidr,
            'ip_version': self._subnet.ip_version,
            'name': self._subnet.name,
            'network_id': self._subnet.network_id,
            'description': self._subnet.description,
        })
        self.assertFalse(self.network.set_tags.called)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def _test_create_with_dns(self, publish_dns=True):
        arglist = [
            "--subnet-range", self._subnet.cidr,
            "--network", self._subnet.network_id,
            self._subnet.name,
        ]
        if publish_dns:
            arglist += ['--dns-publish-fixed-ip']
        else:
            arglist += ['--no-dns-publish-fixed-ip']
        verifylist = [
            ('name', self._subnet.name),
            ('subnet_range', self._subnet.cidr),
            ('network', self._subnet.network_id),
            ('ip_version', self._subnet.ip_version),
            ('gateway', 'auto'),
        ]
        verifylist.append(('dns_publish_fixed_ip', publish_dns))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = (self.cmd.take_action(parsed_args))

        self.network.create_subnet.assert_called_once_with(
            cidr=self._subnet.cidr,
            ip_version=self._subnet.ip_version,
            name=self._subnet.name,
            network_id=self._subnet.network_id,
            dns_publish_fixed_ip=publish_dns,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_dns(self):
        self._test_create_with_dns(publish_dns=True)

    def test_create_with_no_dns(self):
        self._test_create_with_dns(publish_dns=False)

    def _test_create_with_tag(self, add_tags=True):
        arglist = [
            "--subnet-range", self._subnet.cidr,
            "--network", self._subnet.network_id,
            self._subnet.name,
        ]
        if add_tags:
            arglist += ['--tag', 'red', '--tag', 'blue']
        else:
            arglist += ['--no-tag']
        verifylist = [
            ('name', self._subnet.name),
            ('subnet_range', self._subnet.cidr),
            ('network', self._subnet.network_id),
            ('ip_version', self._subnet.ip_version),
            ('gateway', 'auto'),
        ]
        if add_tags:
            verifylist.append(('tags', ['red', 'blue']))
        else:
            verifylist.append(('no_tag', True))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = (self.cmd.take_action(parsed_args))

        self.network.create_subnet.assert_called_once_with(
            cidr=self._subnet.cidr,
            ip_version=self._subnet.ip_version,
            name=self._subnet.name,
            network_id=self._subnet.network_id)
        if add_tags:
            self.network.set_tags.assert_called_once_with(
                self._subnet,
                tests_utils.CompareBySet(['red', 'blue']))
        else:
            self.assertFalse(self.network.set_tags.called)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_create_with_tags(self):
        self._test_create_with_tag(add_tags=True)

    def test_create_with_no_tag(self):
        self._test_create_with_tag(add_tags=False)


class TestDeleteSubnet(TestSubnet):

    # The subnets to delete.
    _subnets = network_fakes.FakeSubnet.create_subnets(count=2)

    def setUp(self):
        super(TestDeleteSubnet, self).setUp()

        self.network.delete_subnet = mock.Mock(return_value=None)

        self.network.find_subnet = (
            network_fakes.FakeSubnet.get_subnets(self._subnets))

        # Get the command object to test
        self.cmd = subnet_v2.DeleteSubnet(self.app, self.namespace)

    def test_subnet_delete(self):
        arglist = [
            self._subnets[0].name,
        ]
        verifylist = [
            ('subnet', [self._subnets[0].name]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)
        self.network.delete_subnet.assert_called_once_with(self._subnets[0])
        self.assertIsNone(result)

    def test_multi_subnets_delete(self):
        arglist = []
        verifylist = []

        for s in self._subnets:
            arglist.append(s.name)
        verifylist = [
            ('subnet', arglist),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        calls = []
        for s in self._subnets:
            calls.append(call(s))
        self.network.delete_subnet.assert_has_calls(calls)
        self.assertIsNone(result)

    def test_multi_subnets_delete_with_exception(self):
        arglist = [
            self._subnets[0].name,
            'unexist_subnet',
        ]
        verifylist = [
            ('subnet',
             [self._subnets[0].name, 'unexist_subnet']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        find_mock_result = [self._subnets[0], exceptions.CommandError]
        self.network.find_subnet = (
            mock.Mock(side_effect=find_mock_result)
        )

        try:
            self.cmd.take_action(parsed_args)
            self.fail('CommandError should be raised.')
        except exceptions.CommandError as e:
            self.assertEqual('1 of 2 subnets failed to delete.', str(e))

        self.network.find_subnet.assert_any_call(
            self._subnets[0].name, ignore_missing=False)
        self.network.find_subnet.assert_any_call(
            'unexist_subnet', ignore_missing=False)
        self.network.delete_subnet.assert_called_once_with(
            self._subnets[0]
        )


class TestListSubnet(TestSubnet):
    # The subnets going to be listed up.
    _subnet = network_fakes.FakeSubnet.create_subnets(count=3)

    columns = (
        'ID',
        'Name',
        'Network',
        'Subnet',
    )
    columns_long = columns + (
        'Project',
        'DHCP',
        'Name Servers',
        'Allocation Pools',
        'Host Routes',
        'IP Version',
        'Gateway',
        'Service Types',
        'Tags',
    )

    data = []
    for subnet in _subnet:
        data.append((
            subnet.id,
            subnet.name,
            subnet.network_id,
            subnet.cidr,
        ))

    data_long = []
    for subnet in _subnet:
        data_long.append((
            subnet.id,
            subnet.name,
            subnet.network_id,
            subnet.cidr,
            subnet.project_id,
            subnet.enable_dhcp,
            format_columns.ListColumn(subnet.dns_nameservers),
            subnet_v2.AllocationPoolsColumn(subnet.allocation_pools),
            subnet_v2.HostRoutesColumn(subnet.host_routes),
            subnet.ip_version,
            subnet.gateway_ip,
            format_columns.ListColumn(subnet.service_types),
            format_columns.ListColumn(subnet.tags),
        ))

    def setUp(self):
        super(TestListSubnet, self).setUp()

        # Get the command object to test
        self.cmd = subnet_v2.ListSubnet(self.app, self.namespace)

        self.network.subnets = mock.Mock(return_value=self._subnet)

    def test_subnet_list_no_options(self):
        arglist = []
        verifylist = [
            ('long', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network.subnets.assert_called_once_with()
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_long(self):
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.network.subnets.assert_called_once_with()
        self.assertEqual(self.columns_long, columns)
        self.assertCountEqual(self.data_long, list(data))

    def test_subnet_list_ip_version(self):
        arglist = [
            '--ip-version', str(4),
        ]
        verifylist = [
            ('ip_version', 4),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'ip_version': 4}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_dhcp(self):
        arglist = [
            '--dhcp',
        ]
        verifylist = [
            ('dhcp', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'enable_dhcp': True, 'is_dhcp_enabled': True}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_no_dhcp(self):
        arglist = [
            '--no-dhcp',
        ]
        verifylist = [
            ('no_dhcp', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'enable_dhcp': False, 'is_dhcp_enabled': False}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_service_type(self):
        arglist = [
            '--service-type', 'network:router_gateway',
        ]
        verifylist = [
            ('service_types', ['network:router_gateway']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        filters = {'service_types': ['network:router_gateway']}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_project(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project', project.id,
        ]
        verifylist = [
            ('project', project.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_service_type_multiple(self):
        arglist = [
            '--service-type', 'network:router_gateway',
            '--service-type', 'network:floatingip_agent_gateway',
        ]
        verifylist = [
            ('service_types', ['network:router_gateway',
                               'network:floatingip_agent_gateway']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'service_types': ['network:router_gateway',
                                     'network:floatingip_agent_gateway']}
        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_project_domain(self):
        project = identity_fakes_v3.FakeProject.create_one_project()
        self.projects_mock.get.return_value = project
        arglist = [
            '--project', project.id,
            '--project-domain', project.domain_id,
        ]
        verifylist = [
            ('project', project.id),
            ('project_domain', project.domain_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'project_id': project.id}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_network(self):
        network = network_fakes.create_one_network()
        self.network.find_network = mock.Mock(return_value=network)
        arglist = [
            '--network', network.id,
        ]
        verifylist = [
            ('network', network.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'network_id': network.id}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_gateway(self):
        subnet = network_fakes.FakeSubnet.create_one_subnet()
        self.network.find_network = mock.Mock(return_value=subnet)
        arglist = [
            '--gateway', subnet.gateway_ip,
        ]
        verifylist = [
            ('gateway', subnet.gateway_ip),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'gateway_ip': subnet.gateway_ip}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_name(self):
        subnet = network_fakes.FakeSubnet.create_one_subnet()
        self.network.find_network = mock.Mock(return_value=subnet)
        arglist = [
            '--name', subnet.name,
        ]
        verifylist = [
            ('name', subnet.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'name': subnet.name}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_subnet_range(self):
        subnet = network_fakes.FakeSubnet.create_one_subnet()
        self.network.find_network = mock.Mock(return_value=subnet)
        arglist = [
            '--subnet-range', subnet.cidr,
        ]
        verifylist = [
            ('subnet_range', subnet.cidr),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'cidr': subnet.cidr}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_subnetpool_by_name(self):
        subnet_pool = network_fakes.FakeSubnetPool.create_one_subnet_pool()
        subnet = network_fakes.FakeSubnet.create_one_subnet(
            {'subnetpool_id': subnet_pool.id})
        self.network.find_network = mock.Mock(return_value=subnet)
        self.network.find_subnet_pool = mock.Mock(return_value=subnet_pool)
        arglist = [
            '--subnet-pool', subnet_pool.name,
        ]
        verifylist = [
            ('subnet_pool', subnet_pool.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'subnetpool_id': subnet_pool.id}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_subnet_list_subnetpool_by_id(self):
        subnet_pool = network_fakes.FakeSubnetPool.create_one_subnet_pool()
        subnet = network_fakes.FakeSubnet.create_one_subnet(
            {'subnetpool_id': subnet_pool.id})
        self.network.find_network = mock.Mock(return_value=subnet)
        self.network.find_subnet_pool = mock.Mock(return_value=subnet_pool)
        arglist = [
            '--subnet-pool', subnet_pool.id,
        ]
        verifylist = [
            ('subnet_pool', subnet_pool.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        filters = {'subnetpool_id': subnet_pool.id}

        self.network.subnets.assert_called_once_with(**filters)
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, list(data))

    def test_list_with_tag_options(self):
        arglist = [
            '--tags', 'red,blue',
            '--any-tags', 'red,green',
            '--not-tags', 'orange,yellow',
            '--not-any-tags', 'black,white',
        ]
        verifylist = [
            ('tags', ['red', 'blue']),
            ('any_tags', ['red', 'green']),
            ('not_tags', ['orange', 'yellow']),
            ('not_any_tags', ['black', 'white']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.subnets.assert_called_once_with(
            **{'tags': 'red,blue',
               'any_tags': 'red,green',
               'not_tags': 'orange,yellow',
               'not_any_tags': 'black,white'}
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestSetSubnet(TestSubnet):

    _subnet = network_fakes.FakeSubnet.create_one_subnet(
        {'tags': ['green', 'red']})

    def setUp(self):
        super(TestSetSubnet, self).setUp()
        self.network.update_subnet = mock.Mock(return_value=None)
        self.network.set_tags = mock.Mock(return_value=None)
        self.network.find_subnet = mock.Mock(return_value=self._subnet)
        self.cmd = subnet_v2.SetSubnet(self.app, self.namespace)

    def test_set_this(self):
        arglist = [
            "--name", "new_subnet",
            "--dhcp",
            "--gateway", self._subnet.gateway_ip,
            self._subnet.name,
        ]
        verifylist = [
            ('name', "new_subnet"),
            ('dhcp', True),
            ('gateway', self._subnet.gateway_ip),
            ('subnet', self._subnet.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'enable_dhcp': True,
            'gateway_ip': self._subnet.gateway_ip,
            'name': "new_subnet",
        }
        self.network.update_subnet.assert_called_with(self._subnet, **attrs)
        self.assertIsNone(result)

    def test_set_that(self):
        arglist = [
            "--name", "new_subnet",
            "--no-dhcp",
            "--gateway", "none",
            self._subnet.name,
        ]
        verifylist = [
            ('name', "new_subnet"),
            ('no_dhcp', True),
            ('gateway', "none"),
            ('subnet', self._subnet.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'enable_dhcp': False,
            'gateway_ip': None,
            'name': "new_subnet",
        }
        self.network.update_subnet.assert_called_with(self._subnet, **attrs)
        self.assertIsNone(result)

    def test_set_nothing(self):
        arglist = [self._subnet.name, ]
        verifylist = [('subnet', self._subnet.name)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network.update_subnet.called)
        self.assertFalse(self.network.set_tags.called)
        self.assertIsNone(result)

    def test_append_options(self):
        _testsubnet = network_fakes.FakeSubnet.create_one_subnet(
            {'dns_nameservers': ["10.0.0.1"],
             'service_types': ["network:router_gateway"]})
        self.network.find_subnet = mock.Mock(return_value=_testsubnet)
        arglist = [
            '--dns-nameserver', '10.0.0.2',
            '--service-type', 'network:floatingip_agent_gateway',
            _testsubnet.name,
        ]
        verifylist = [
            ('dns_nameservers', ['10.0.0.2']),
            ('service_types', ['network:floatingip_agent_gateway']),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'dns_nameservers': ['10.0.0.2', '10.0.0.1'],
            'service_types': ['network:floatingip_agent_gateway',
                              'network:router_gateway'],
        }
        self.network.update_subnet.assert_called_once_with(
            _testsubnet, **attrs)
        self.assertIsNone(result)

    def test_set_non_append_options(self):
        arglist = [
            "--description", "new_description",
            "--dhcp",
            "--gateway", self._subnet.gateway_ip,
            self._subnet.name,
        ]
        verifylist = [
            ('description', "new_description"),
            ('dhcp', True),
            ('gateway', self._subnet.gateway_ip),
            ('subnet', self._subnet.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'enable_dhcp': True,
            'gateway_ip': self._subnet.gateway_ip,
            'description': "new_description",
        }
        self.network.update_subnet.assert_called_with(self._subnet, **attrs)
        self.assertIsNone(result)

    def test_overwrite_options(self):
        _testsubnet = network_fakes.FakeSubnet.create_one_subnet(
            {'host_routes': [{'destination': '10.20.20.0/24',
                              'nexthop': '10.20.20.1'}],
             'allocation_pools': [{'start': '8.8.8.200',
                                   'end': '8.8.8.250'}],
             'dns_nameservers': ["10.0.0.1"], })
        self.network.find_subnet = mock.Mock(return_value=_testsubnet)
        arglist = [
            '--host-route', 'destination=10.30.30.30/24,gateway=10.30.30.1',
            '--no-host-route',
            '--allocation-pool', 'start=8.8.8.100,end=8.8.8.150',
            '--no-allocation-pool',
            '--dns-nameserver', '10.1.10.1',
            '--no-dns-nameservers',
            _testsubnet.name,
        ]
        verifylist = [
            ('host_routes', [{
                "destination": "10.30.30.30/24", "gateway": "10.30.30.1"}]),
            ('allocation_pools', [{
                'start': '8.8.8.100', 'end': '8.8.8.150'}]),
            ('dns_nameservers', ['10.1.10.1']),
            ('no_dns_nameservers', True),
            ('no_host_route', True),
            ('no_allocation_pool', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'host_routes': [{
                "destination": "10.30.30.30/24", "nexthop": "10.30.30.1"}],
            'allocation_pools': [{'start': '8.8.8.100', 'end': '8.8.8.150'}],
            'dns_nameservers': ["10.1.10.1"],
        }
        self.network.update_subnet.assert_called_once_with(
            _testsubnet, **attrs)
        self.assertIsNone(result)

    def test_clear_options(self):
        _testsubnet = network_fakes.FakeSubnet.create_one_subnet(
            {'host_routes': [{'destination': '10.20.20.0/24',
                              'nexthop': '10.20.20.1'}],
             'allocation_pools': [{'start': '8.8.8.200',
                                   'end': '8.8.8.250'}],
             'dns_nameservers': ['10.0.0.1'], })
        self.network.find_subnet = mock.Mock(return_value=_testsubnet)
        arglist = [
            '--no-host-route',
            '--no-allocation-pool',
            '--no-dns-nameservers',
            _testsubnet.name,
        ]
        verifylist = [
            ('no_dns_nameservers', True),
            ('no_host_route', True),
            ('no_allocation_pool', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {
            'host_routes': [],
            'allocation_pools': [],
            'dns_nameservers': [],
        }
        self.network.update_subnet.assert_called_once_with(
            _testsubnet, **attrs)
        self.assertIsNone(result)

    def _test_set_tags(self, with_tags=True):
        if with_tags:
            arglist = ['--tag', 'red', '--tag', 'blue']
            verifylist = [('tags', ['red', 'blue'])]
            expected_args = ['red', 'blue', 'green']
        else:
            arglist = ['--no-tag']
            verifylist = [('no_tag', True)]
            expected_args = []
        arglist.append(self._subnet.name)
        verifylist.append(
            ('subnet', self._subnet.name))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network.update_subnet.called)
        self.network.set_tags.assert_called_once_with(
            self._subnet,
            tests_utils.CompareBySet(expected_args))
        self.assertIsNone(result)

    def test_set_with_tags(self):
        self._test_set_tags(with_tags=True)

    def test_set_with_no_tag(self):
        self._test_set_tags(with_tags=False)

    def test_set_segment(self):
        _net = network_fakes.create_one_network()
        _segment = network_fakes.create_one_network_segment(
            attrs={'network_id': _net.id})
        _subnet = network_fakes.FakeSubnet.create_one_subnet(
            {'host_routes': [{'destination': '10.20.20.0/24',
                              'nexthop': '10.20.20.1'}],
             'allocation_pools': [{'start': '8.8.8.200',
                                   'end': '8.8.8.250'}],
             'dns_nameservers': ["10.0.0.1"],
             'network_id': _net.id,
             'segment_id': None})
        self.network.find_subnet = mock.Mock(return_value=_subnet)
        self.network.find_segment = mock.Mock(return_value=_segment)
        arglist = ['--network-segment', _segment.id, _subnet.name]
        verifylist = [('network_segment', _segment.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)
        attrs = {'segment_id': _segment.id}
        self.network.update_subnet.assert_called_once_with(_subnet, **attrs)
        self.network.update_subnet.assert_called_with(_subnet, **attrs)
        self.assertIsNone(result)


class TestShowSubnet(TestSubnet):
    # The subnets to be shown
    _subnet = network_fakes.FakeSubnet.create_one_subnet()

    columns = (
        'allocation_pools',
        'cidr',
        'description',
        'dns_nameservers',
        'enable_dhcp',
        'gateway_ip',
        'host_routes',
        'id',
        'ip_version',
        'ipv6_address_mode',
        'ipv6_ra_mode',
        'name',
        'network_id',
        'project_id',
        'segment_id',
        'service_types',
        'subnetpool_id',
        'tags',
    )

    data = (
        subnet_v2.AllocationPoolsColumn(_subnet.allocation_pools),
        _subnet.cidr,
        _subnet.description,
        format_columns.ListColumn(_subnet.dns_nameservers),
        _subnet.enable_dhcp,
        _subnet.gateway_ip,
        subnet_v2.HostRoutesColumn(_subnet.host_routes),
        _subnet.id,
        _subnet.ip_version,
        _subnet.ipv6_address_mode,
        _subnet.ipv6_ra_mode,
        _subnet.name,
        _subnet.network_id,
        _subnet.project_id,
        _subnet.segment_id,
        format_columns.ListColumn(_subnet.service_types),
        _subnet.subnetpool_id,
        format_columns.ListColumn(_subnet.tags),
    )

    def setUp(self):
        super(TestShowSubnet, self).setUp()

        # Get the command object to test
        self.cmd = subnet_v2.ShowSubnet(self.app, self.namespace)

        self.network.find_subnet = mock.Mock(return_value=self._subnet)

    def test_show_no_options(self):
        arglist = []
        verifylist = []

        # Testing that a call without the required argument will fail and
        # throw a "ParserExecption"
        self.assertRaises(tests_utils.ParserException,
                          self.check_parser, self.cmd, arglist, verifylist)

    def test_show_all_options(self):
        arglist = [
            self._subnet.name,
        ]
        verifylist = [
            ('subnet', self._subnet.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.network.find_subnet.assert_called_once_with(
            self._subnet.name, ignore_missing=False)

        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)


class TestUnsetSubnet(TestSubnet):

    def setUp(self):
        super(TestUnsetSubnet, self).setUp()
        self._testsubnet = network_fakes.FakeSubnet.create_one_subnet(
            {'dns_nameservers': ['8.8.8.8',
                                 '8.8.8.4'],
             'host_routes': [{'destination': '10.20.20.0/24',
                              'nexthop': '10.20.20.1'},
                             {'destination': '10.30.30.30/24',
                              'nexthop': '10.30.30.1'}],
             'allocation_pools': [{'start': '8.8.8.100',
                                   'end': '8.8.8.150'},
                                  {'start': '8.8.8.160',
                                   'end': '8.8.8.170'}],
             'service_types': ['network:router_gateway',
                               'network:floatingip_agent_gateway'],
             'gateway_ip': 'fe80::a00a:0:c0de:0:1',
             'tags': ['green', 'red'], })
        self.network.find_subnet = mock.Mock(return_value=self._testsubnet)
        self.network.update_subnet = mock.Mock(return_value=None)
        self.network.set_tags = mock.Mock(return_value=None)
        # Get the command object to test
        self.cmd = subnet_v2.UnsetSubnet(self.app, self.namespace)

    def test_unset_subnet_params(self):
        arglist = [
            '--dns-nameserver', '8.8.8.8',
            '--host-route', 'destination=10.30.30.30/24,gateway=10.30.30.1',
            '--allocation-pool', 'start=8.8.8.100,end=8.8.8.150',
            '--service-type', 'network:router_gateway',
            '--gateway',
            self._testsubnet.name,
        ]
        verifylist = [
            ('dns_nameservers', ['8.8.8.8']),
            ('host_routes', [{
                "destination": "10.30.30.30/24", "gateway": "10.30.30.1"}]),
            ('allocation_pools', [{
                'start': '8.8.8.100', 'end': '8.8.8.150'}]),
            ('service_types', ['network:router_gateway']),
            ('gateway', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        attrs = {
            'dns_nameservers': ['8.8.8.4'],
            'host_routes': [{
                "destination": "10.20.20.0/24", "nexthop": "10.20.20.1"}],
            'allocation_pools': [{'start': '8.8.8.160', 'end': '8.8.8.170'}],
            'service_types': ['network:floatingip_agent_gateway'],
            'gateway_ip': None,
        }
        self.network.update_subnet.assert_called_once_with(
            self._testsubnet, **attrs)
        self.assertIsNone(result)

    def test_unset_subnet_wrong_host_routes(self):
        arglist = [
            '--dns-nameserver', '8.8.8.8',
            '--host-route', 'destination=10.30.30.30/24,gateway=10.30.30.2',
            '--allocation-pool', 'start=8.8.8.100,end=8.8.8.150',
            self._testsubnet.name,
        ]
        verifylist = [
            ('dns_nameservers', ['8.8.8.8']),
            ('host_routes', [{
                "destination": "10.30.30.30/24", "gateway": "10.30.30.2"}]),
            ('allocation_pools', [{
                'start': '8.8.8.100', 'end': '8.8.8.150'}]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action, parsed_args)

    def test_unset_subnet_wrong_allocation_pool(self):
        arglist = [
            '--dns-nameserver', '8.8.8.8',
            '--host-route', 'destination=10.30.30.30/24,gateway=10.30.30.1',
            '--allocation-pool', 'start=8.8.8.100,end=8.8.8.156',
            self._testsubnet.name,
        ]
        verifylist = [
            ('dns_nameservers', ['8.8.8.8']),
            ('host_routes', [{
                "destination": "10.30.30.30/24", "gateway": "10.30.30.1"}]),
            ('allocation_pools', [{
                'start': '8.8.8.100', 'end': '8.8.8.156'}]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action, parsed_args)

    def test_unset_subnet_wrong_dns_nameservers(self):
        arglist = [
            '--dns-nameserver', '8.8.8.1',
            '--host-route', 'destination=10.30.30.30/24,gateway=10.30.30.1',
            '--allocation-pool', 'start=8.8.8.100,end=8.8.8.150',
            self._testsubnet.name,
        ]
        verifylist = [
            ('dns_nameservers', ['8.8.8.1']),
            ('host_routes', [{
                "destination": "10.30.30.30/24", "gateway": "10.30.30.1"}]),
            ('allocation_pools', [{
                'start': '8.8.8.100', 'end': '8.8.8.150'}]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action, parsed_args)

    def test_unset_subnet_wrong_service_type(self):
        arglist = [
            '--dns-nameserver', '8.8.8.8',
            '--host-route', 'destination=10.30.30.30/24,gateway=10.30.30.1',
            '--allocation-pool', 'start=8.8.8.100,end=8.8.8.150',
            '--service-type', 'network:dhcp',
            self._testsubnet.name,
        ]
        verifylist = [
            ('dns_nameservers', ['8.8.8.8']),
            ('host_routes', [{
                "destination": "10.30.30.30/24", "gateway": "10.30.30.1"}]),
            ('allocation_pools', [{
                'start': '8.8.8.100', 'end': '8.8.8.150'}]),
            ('service_types', ['network:dhcp']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exceptions.CommandError,
                          self.cmd.take_action, parsed_args)

    def _test_unset_tags(self, with_tags=True):
        if with_tags:
            arglist = ['--tag', 'red', '--tag', 'blue']
            verifylist = [('tags', ['red', 'blue'])]
            expected_args = ['green']
        else:
            arglist = ['--all-tag']
            verifylist = [('all_tag', True)]
            expected_args = []
        arglist.append(self._testsubnet.name)
        verifylist.append(
            ('subnet', self._testsubnet.name))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        self.assertFalse(self.network.update_subnet.called)
        self.network.set_tags.assert_called_once_with(
            self._testsubnet,
            tests_utils.CompareBySet(expected_args))
        self.assertIsNone(result)

    def test_unset_with_tags(self):
        self._test_unset_tags(with_tags=True)

    def test_unset_with_all_tag(self):
        self._test_unset_tags(with_tags=False)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v3 import common


class ProjectTests(common.IdentityTests):

    def test_project_create(self):
        project_name = data_utils.rand_name('TestProject')
        description = data_utils.rand_name('description')
        raw_output = self.openstack(
            'project create '
            '--domain %(domain)s '
            '--description %(description)s '
            '--enable '
            '--property k1=v1 '
            '--property k2=v2 '
            '%(name)s' % {'domain': self.domain_name,
                          'description': description,
                          'name': project_name})
        self.addCleanup(
            self.openstack,
            'project delete '
            '--domain %(domain)s '
            '%(name)s' % {'domain': self.domain_name,
                          'name': project_name}
        )
        items = self.parse_show(raw_output)
        show_fields = list(self.PROJECT_FIELDS)
        show_fields.extend(['k1', 'k2'])
        self.assert_show_fields(items, show_fields)
        project = self.parse_show_as_object(raw_output)
        self.assertEqual('v1', project['k1'])
        self.assertEqual('v2', project['k2'])

    def test_project_delete(self):
        project_name = self._create_dummy_project(add_clean_up=False)
        raw_output = self.openstack(
            'project delete '
            '--domain %(domain)s '
            '%(name)s' % {'domain': self.domain_name,
                          'name': project_name})
        self.assertEqual(0, len(raw_output))

    def test_project_list(self):
        raw_output = self.openstack('project list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, common.BASIC_LIST_HEADERS)

    def test_project_list_with_domain(self):
        project_name = self._create_dummy_project()
        raw_output = self.openstack(
            'project list --domain %s' % self.domain_name)
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, common.BASIC_LIST_HEADERS)
        self.assertIn(project_name, raw_output)
        self.assertGreater(len(items), 0)

    def test_project_set(self):
        project_name = self._create_dummy_project()
        new_project_name = data_utils.rand_name('NewTestProject')
        raw_output = self.openstack(
            'project set '
            '--name %(new_name)s '
            '--disable '
            '--property k0=v0 '
            '%(name)s' % {'new_name': new_project_name,
                          'name': project_name})
        self.assertEqual(0, len(raw_output))
        # check project details
        raw_output = self.openstack(
            'project show '
            '--domain %(domain)s '
            '%(name)s' % {'domain': self.domain_name,
                          'name': new_project_name}
        )
        items = self.parse_show(raw_output)
        fields = list(self.PROJECT_FIELDS)
        fields.extend(['k0'])
        self.assert_show_fields(items, fields)
        project = self.parse_show_as_object(raw_output)
        self.assertEqual(new_project_name, project['name'])
        self.assertEqual('False', project['enabled'])
        self.assertEqual('v0', project['k0'])
        # reset project to make sure it will be cleaned up
        self.openstack(
            'project set '
            '--name %(new_name)s '
            '--enable '
            '%(name)s' % {'new_name': project_name,
                          'name': new_project_name})

    def test_project_show(self):
        raw_output = self.openstack(
            'project show '
            '--domain %(domain)s '
            '%(name)s' % {'domain': self.domain_name,
                          'name': self.project_name})
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.PROJECT_FIELDS)

    def test_project_show_with_parents_children(self):
        output = self.openstack(
            'project show '
            '--parents --children '
            '--domain %(domain)s '
            '%(name)s' % {'domain': self.domain_name,
                          'name': self.project_name},
            parse_output=True,
        )
        for attr_name in (self.PROJECT_FIELDS + ['parents', 'subtree']):
            self.assertIn(attr_name, output)
        self.assertEqual(self.project_name, output.get('name'))



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Limits Action Implementation"""

import itertools

from osc_lib.command import command
from osc_lib import utils

from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


class ShowLimits(command.Lister):
    _description = _("Show compute and block storage limits")

    def get_parser(self, prog_name):
        parser = super(ShowLimits, self).get_parser(prog_name)
        type_group = parser.add_mutually_exclusive_group(required=True)
        type_group.add_argument(
            "--absolute",
            dest="is_absolute",
            action="store_true",
            default=False,
            help=_("Show absolute limits"),
        )
        type_group.add_argument(
            "--rate",
            dest="is_rate",
            action="store_true",
            default=False,
            help=_("Show rate limits"),
        )
        parser.add_argument(
            "--reserved",
            dest="is_reserved",
            action="store_true",
            default=False,
            help=_("Include reservations count [only valid with --absolute]"),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_('Show limits for a specific project (name or ID)'
                   ' [only valid with --absolute]'),
        )
        parser.add_argument(
            '--domain',
            metavar='<domain>',
            help=_('Domain the project belongs to (name or ID)'
                   ' [only valid with --absolute]'),
        )
        return parser

    def take_action(self, parsed_args):

        compute_client = self.app.client_manager.compute
        volume_client = self.app.client_manager.volume

        project_id = None
        if parsed_args.project is not None:
            identity_client = self.app.client_manager.identity
            if parsed_args.domain is not None:
                domain = identity_common.find_domain(identity_client,
                                                     parsed_args.domain)
                project_id = utils.find_resource(identity_client.projects,
                                                 parsed_args.project,
                                                 domain_id=domain.id).id
            else:
                project_id = utils.find_resource(identity_client.projects,
                                                 parsed_args.project).id

        compute_limits = None
        volume_limits = None

        if self.app.client_manager.is_compute_endpoint_enabled():
            compute_limits = compute_client.limits.get(parsed_args.is_reserved,
                                                       tenant_id=project_id)

        if self.app.client_manager.is_volume_endpoint_enabled(volume_client):
            volume_limits = volume_client.limits.get()

        data = []
        if parsed_args.is_absolute:
            if compute_limits:
                data.append(compute_limits.absolute)
            if volume_limits:
                data.append(volume_limits.absolute)
            columns = ["Name", "Value"]
            return (columns, (utils.get_item_properties(s, columns)
                              for s in itertools.chain(*data)))

        elif parsed_args.is_rate:
            if compute_limits:
                data.append(compute_limits.rate)
            if volume_limits:
                data.append(volume_limits.rate)
            columns = ["Verb", "URI", "Value", "Remain", "Unit",
                       "Next Available"]
            return (columns, (utils.get_item_properties(s, columns)
                              for s in itertools.chain(*data)))
        else:
            return {}, {}



#   Copyright 2017 Huawei, Inc. All rights reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock

import iso8601
from novaclient import api_versions
from osc_lib import exceptions

from openstackclient.compute.v2 import server_event
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes


class TestServerEvent(compute_fakes.TestComputev2):

    fake_server = compute_fakes.FakeServer.create_one_server()

    def setUp(self):
        super(TestServerEvent, self).setUp()

        self.servers_mock = self.app.client_manager.compute.servers
        self.servers_mock.reset_mock()
        self.events_mock = self.app.client_manager.compute.instance_action
        self.events_mock.reset_mock()

        self.servers_mock.get.return_value = self.fake_server


class TestListServerEvent(TestServerEvent):

    fake_event = compute_fakes.FakeServerEvent.create_one_server_event()

    columns = (
        'Request ID',
        'Server ID',
        'Action',
        'Start Time',
    )
    data = ((
        fake_event.request_id,
        fake_event.instance_uuid,
        fake_event.action,
        fake_event.start_time,
    ), )

    long_columns = (
        'Request ID',
        'Server ID',
        'Action',
        'Start Time',
        'Message',
        'Project ID',
        'User ID',
    )
    long_data = ((
        fake_event.request_id,
        fake_event.instance_uuid,
        fake_event.action,
        fake_event.start_time,
        fake_event.message,
        fake_event.project_id,
        fake_event.user_id,
    ), )

    def setUp(self):
        super(TestListServerEvent, self).setUp()

        self.events_mock.list.return_value = [self.fake_event, ]
        self.cmd = server_event.ListServerEvent(self.app, None)

    def test_server_event_list(self):
        arglist = [
            self.fake_server.name,
        ]
        verifylist = [
            ('server', self.fake_server.name),
            ('long', False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.servers_mock.get.assert_called_once_with(self.fake_server.name)
        self.events_mock.list.assert_called_once_with(self.fake_server.id)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, tuple(data))

    def test_server_event_list_long(self):
        arglist = [
            '--long',
            self.fake_server.name,
        ]
        verifylist = [
            ('server', self.fake_server.name),
            ('long', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.servers_mock.get.assert_called_once_with(self.fake_server.name)
        self.events_mock.list.assert_called_once_with(self.fake_server.id)

        self.assertEqual(self.long_columns, columns)
        self.assertEqual(self.long_data, tuple(data))

    def test_server_event_list_with_changes_since(self):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.58')

        arglist = [
            '--changes-since', '2016-03-04T06:27:59Z',
            self.fake_server.name,
        ]
        verifylist = [
            ('server', self.fake_server.name),
            ('changes_since', '2016-03-04T06:27:59Z'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.servers_mock.get.assert_called_once_with(self.fake_server.name)
        self.events_mock.list.assert_called_once_with(
            self.fake_server.id, changes_since='2016-03-04T06:27:59Z')

        self.assertEqual(self.columns, columns)
        self.assertEqual(tuple(self.data), tuple(data))

    @mock.patch.object(iso8601, 'parse_date', side_effect=iso8601.ParseError)
    def test_server_event_list_with_changes_since_invalid(
        self, mock_parse_isotime,
    ):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.58')

        arglist = [
            '--changes-since', 'Invalid time value',
            self.fake_server.name,
        ]
        verifylist = [
            ('server', self.fake_server.name),
            ('changes_since', 'Invalid time value'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)

        self.assertIn(
            'Invalid changes-since value:', str(ex))
        mock_parse_isotime.assert_called_once_with(
            'Invalid time value'
        )

    def test_server_event_list_with_changes_since_pre_v258(self):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.57')

        arglist = [
            '--changes-since', '2016-03-04T06:27:59Z',
            self.fake_server.name,
        ]
        verifylist = [
            ('server', self.fake_server.name),
            ('changes_since', '2016-03-04T06:27:59Z'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)

        self.assertIn(
            '--os-compute-api-version 2.58 or greater is required', str(ex))

    def test_server_event_list_with_changes_before(self):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.66')

        arglist = [
            '--changes-before', '2016-03-04T06:27:59Z',
            self.fake_server.name,
        ]
        verifylist = [
            ('server', self.fake_server.name),
            ('changes_before', '2016-03-04T06:27:59Z'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.servers_mock.get.assert_called_once_with(self.fake_server.name)
        self.events_mock.list.assert_called_once_with(
            self.fake_server.id, changes_before='2016-03-04T06:27:59Z')

        self.assertEqual(self.columns, columns)
        self.assertEqual(tuple(self.data), tuple(data))

    @mock.patch.object(iso8601, 'parse_date', side_effect=iso8601.ParseError)
    def test_server_event_list_with_changes_before_invalid(
        self, mock_parse_isotime,
    ):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.66')

        arglist = [
            '--changes-before', 'Invalid time value',
            self.fake_server.name,
        ]
        verifylist = [
            ('server', self.fake_server.name),
            ('changes_before', 'Invalid time value'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)

        self.assertIn(
            'Invalid changes-before value:', str(ex))
        mock_parse_isotime.assert_called_once_with(
            'Invalid time value'
        )

    def test_server_event_list_with_changes_before_pre_v266(self):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.65')

        arglist = [
            '--changes-before', '2016-03-04T06:27:59Z',
            self.fake_server.name,
        ]
        verifylist = [
            ('server', self.fake_server.name),
            ('changes_before', '2016-03-04T06:27:59Z'),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)

        self.assertIn(
            '--os-compute-api-version 2.66 or greater is required', str(ex))

    def test_server_event_list_with_limit(self):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.58')

        arglist = [
            '--limit', '1',
            self.fake_server.name,
        ]
        verifylist = [
            ('limit', 1),
            ('server', self.fake_server.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.events_mock.list.assert_called_once_with(
            self.fake_server.id, limit=1)

    def test_server_event_list_with_limit_pre_v258(self):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.57')

        arglist = [
            '--limit', '1',
            self.fake_server.name,
        ]
        verifylist = [
            ('limit', 1),
            ('server', self.fake_server.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)

        self.assertIn(
            '--os-compute-api-version 2.58 or greater is required', str(ex))

    def test_server_event_list_with_marker(self):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.58')

        arglist = [
            '--marker', 'test_event',
            self.fake_server.name,
        ]
        verifylist = [
            ('marker', 'test_event'),
            ('server', self.fake_server.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.events_mock.list.assert_called_once_with(
            self.fake_server.id, marker='test_event')

    def test_server_event_list_with_marker_pre_v258(self):
        self.app.client_manager.compute.api_version = \
            api_versions.APIVersion('2.57')

        arglist = [
            '--marker', 'test_event',
            self.fake_server.name,
        ]
        verifylist = [
            ('marker', 'test_event'),
            ('server', self.fake_server.name),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        ex = self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args)

        self.assertIn(
            '--os-compute-api-version 2.58 or greater is required', str(ex))


class TestShowServerEvent(TestServerEvent):

    fake_event = compute_fakes.FakeServerEvent.create_one_server_event()

    columns = (
        'action',
        'events',
        'instance_uuid',
        'message',
        'project_id',
        'request_id',
        'start_time',
        'user_id',
    )
    data = (
        fake_event.action,
        fake_event.events,
        fake_event.instance_uuid,
        fake_event.message,
        fake_event.project_id,
        fake_event.request_id,
        fake_event.start_time,
        fake_event.user_id,
    )

    def setUp(self):
        super(TestShowServerEvent, self).setUp()

        self.events_mock.get.return_value = self.fake_event
        self.cmd = server_event.ShowServerEvent(self.app, None)

    def test_server_event_show(self):
        arglist = [
            self.fake_server.name,
            self.fake_event.request_id,
        ]
        verifylist = [
            ('server', self.fake_server.name),
            ('request_id', self.fake_event.request_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.servers_mock.get.assert_called_once_with(self.fake_server.name)
        self.events_mock.get.assert_called_once_with(
            self.fake_server.id, self.fake_event.request_id)

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from keystoneauth1 import fixture as ksa_fixture
from requests_mock.contrib import fixture

from openstackclient.tests.unit import test_shell
from openstackclient.tests.unit import utils


HOST = "192.168.5.41"
URL_BASE = "http://%s/identity" % HOST

V2_AUTH_URL = URL_BASE + "/v2.0/"
V2_VERSION_RESP = {
    "version": {
        "status": "stable",
        "updated": "2014-04-17T00:00:00Z",
        "media-types": [
            {
                "base": "application/json",
                "type": "application/vnd.openstack.identity-v2.0+json",
            },
        ],
        "id": "v2.0",
        "links": [
            {
                "href": V2_AUTH_URL,
                "rel": "self",
            },
            {
                "href": "http://docs.openstack.org/",
                "type": "text/html",
                "rel": "describedby",
            },
        ],
    },
}

V3_AUTH_URL = URL_BASE + "/v3/"
V3_VERSION_RESP = {
    "version": {
        "status": "stable",
        "updated": "2016-04-04T00:00:00Z",
        "media-types": [{
            "base": "application/json",
            "type": "application/vnd.openstack.identity-v3+json",
        }],
        "id": "v3.6",
        "links": [{
            "href": V3_AUTH_URL,
            "rel": "self",
        }]
    }
}


def make_v2_token(req_mock):
    """Create an Identity v2 token and register the responses"""

    token = ksa_fixture.V2Token(
        tenant_name=test_shell.DEFAULT_PROJECT_NAME,
        user_name=test_shell.DEFAULT_USERNAME,
    )

    # Set up the v2 auth routes
    req_mock.register_uri(
        'GET',
        V2_AUTH_URL,
        json=V2_VERSION_RESP,
        status_code=200,
    )
    req_mock.register_uri(
        'POST',
        V2_AUTH_URL + 'tokens',
        json=token,
        status_code=200,
    )
    return token


def make_v3_token(req_mock):
    """Create an Identity v3 token and register the response"""

    token = ksa_fixture.V3Token(
        # project_domain_id=test_shell.DEFAULT_PROJECT_DOMAIN_ID,
        user_domain_id=test_shell.DEFAULT_USER_DOMAIN_ID,
        user_name=test_shell.DEFAULT_USERNAME,
    )

    # Set up the v3 auth routes
    req_mock.register_uri(
        'GET',
        V3_AUTH_URL,
        json=V3_VERSION_RESP,
        status_code=200,
    )
    req_mock.register_uri(
        'POST',
        V3_AUTH_URL + 'auth/tokens',
        json=token,
        status_code=200,
    )
    return token


class TestInteg(utils.TestCase):

    def setUp(self):
        super(TestInteg, self).setUp()

        self.requests_mock = self.useFixture(fixture.Fixture())



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Volume v1 Backup action implementations"""

import copy
import functools
import logging

from cliff import columns as cliff_columns
from osc_lib.command import command
from osc_lib import exceptions
from osc_lib import utils

from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


class VolumeIdColumn(cliff_columns.FormattableColumn):
    """Formattable column for volume ID column.

    Unlike the parent FormattableColumn class, the initializer of the
    class takes volume_cache as the second argument.
    osc_lib.utils.get_item_properties instantiate cliff FormattableColumn
    object with a single parameter "column value", so you need to pass
    a partially initialized class like
    ``functools.partial(VolumeIdColumn, volume_cache)``.
    """

    def __init__(self, value, volume_cache=None):
        super(VolumeIdColumn, self).__init__(value)
        self._volume_cache = volume_cache or {}

    def human_readable(self):
        """Return a volume name if available

        :rtype: either the volume ID or name
        """
        volume_id = self._value
        volume = volume_id
        if volume_id in self._volume_cache.keys():
            volume = self._volume_cache[volume_id].display_name
        return volume


class CreateVolumeBackup(command.ShowOne):
    _description = _("Create new volume backup")

    def get_parser(self, prog_name):
        parser = super(CreateVolumeBackup, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=_('Volume to backup (name or ID)'),
        )
        parser.add_argument(
            '--container',
            metavar='<container>',
            required=False,
            help=_('Optional backup container name'),
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Name of the backup'),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description of the backup'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        volume_id = utils.find_resource(volume_client.volumes,
                                        parsed_args.volume).id
        backup = volume_client.backups.create(
            volume_id,
            parsed_args.container,
            parsed_args.name,
            parsed_args.description
        )

        backup._info.pop('links')
        return zip(*sorted(backup._info.items()))


class DeleteVolumeBackup(command.Command):
    _description = _("Delete volume backup(s)")

    def get_parser(self, prog_name):
        parser = super(DeleteVolumeBackup, self).get_parser(prog_name)
        parser.add_argument(
            'backups',
            metavar='<backup>',
            nargs="+",
            help=_('Backup(s) to delete (name or ID)'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        result = 0

        for i in parsed_args.backups:
            try:
                backup_id = utils.find_resource(
                    volume_client.backups, i).id
                volume_client.backups.delete(backup_id)
            except Exception as e:
                result += 1
                LOG.error(_("Failed to delete backup with "
                            "name or ID '%(backup)s': %(e)s"),
                          {'backup': i, 'e': e})

        if result > 0:
            total = len(parsed_args.backups)
            msg = (_("%(result)s of %(total)s backups failed "
                   "to delete.") % {'result': result, 'total': total})
            raise exceptions.CommandError(msg)


class ListVolumeBackup(command.Lister):
    _description = _("List volume backups")

    def get_parser(self, prog_name):
        parser = super(ListVolumeBackup, self).get_parser(prog_name)
        parser.add_argument(
            '--long',
            action='store_true',
            default=False,
            help=_('List additional fields in output'),
        )
        parser.add_argument(
            "--name",
            metavar="<name>",
            help=_("Filters results by the backup name")
        )
        parser.add_argument(
            "--status",
            metavar="<status>",
            choices=['creating', 'available', 'deleting',
                     'error', 'restoring', 'error_restoring'],
            help=_("Filters results by the backup status "
                   "('creating', 'available', 'deleting', "
                   "'error', 'restoring' or 'error_restoring')")
        )
        parser.add_argument(
            "--volume",
            metavar="<volume>",
            help=_("Filters results by the volume which they "
                   "backup (name or ID)")
        )
        parser.add_argument(
            '--all-projects',
            action='store_true',
            default=False,
            help=_('Include all projects (admin only)'),
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume

        if parsed_args.long:
            columns = ['ID', 'Name', 'Description', 'Status', 'Size',
                       'Availability Zone', 'Volume ID', 'Container']
            column_headers = copy.deepcopy(columns)
            column_headers[6] = 'Volume'
        else:
            columns = ['ID', 'Name', 'Description', 'Status', 'Size']
            column_headers = columns

        # Cache the volume list
        volume_cache = {}
        try:
            for s in volume_client.volumes.list():
                volume_cache[s.id] = s
        except Exception:
            # Just forget it if there's any trouble
            pass
        VolumeIdColumnWithCache = functools.partial(VolumeIdColumn,
                                                    volume_cache=volume_cache)

        filter_volume_id = None
        if parsed_args.volume:
            filter_volume_id = utils.find_resource(volume_client.volumes,
                                                   parsed_args.volume).id
        search_opts = {
            'name': parsed_args.name,
            'status': parsed_args.status,
            'volume_id': filter_volume_id,
            'all_tenants': parsed_args.all_projects,
        }
        data = volume_client.backups.list(
            search_opts=search_opts,
        )

        return (column_headers,
                (utils.get_item_properties(
                    s, columns,
                    formatters={'Volume ID': VolumeIdColumnWithCache},
                ) for s in data))


class RestoreVolumeBackup(command.Command):
    _description = _("Restore volume backup")

    def get_parser(self, prog_name):
        parser = super(RestoreVolumeBackup, self).get_parser(prog_name)
        parser.add_argument(
            'backup',
            metavar='<backup>',
            help=_('Backup to restore (name or ID)')
        )
        parser.add_argument(
            'volume',
            metavar='<volume>',
            nargs='?',
            help=_('Volume to restore to (name or ID) (default to None)')
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        backup = utils.find_resource(
            volume_client.backups, parsed_args.backup,
        )
        volume_id = None
        if parsed_args.volume is not None:
            volume_id = utils.find_resource(
                volume_client.volumes,
                parsed_args.volume,
            ).id
        return volume_client.restores.restore(backup.id, volume_id)


class ShowVolumeBackup(command.ShowOne):
    _description = _("Display volume backup details")

    def get_parser(self, prog_name):
        parser = super(ShowVolumeBackup, self).get_parser(prog_name)
        parser.add_argument(
            'backup',
            metavar='<backup>',
            help=_('Backup to display (name or ID)')
        )
        return parser

    def take_action(self, parsed_args):
        volume_client = self.app.client_manager.volume
        backup = utils.find_resource(volume_client.backups,
                                     parsed_args.backup)
        backup._info.pop('links')
        return zip(*sorted(backup._info.items()))



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import random
from unittest import mock
import uuid

from openstack.image.v2 import image
from openstack.image.v2 import member
from openstack.image.v2 import metadef_namespace
from openstack.image.v2 import service_info as _service_info
from openstack.image.v2 import task

from openstackclient.tests.unit import fakes
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit import utils


class FakeImagev2Client:

    def __init__(self, **kwargs):
        self.images = mock.Mock()
        self.create_image = mock.Mock()
        self.delete_image = mock.Mock()
        self.update_image = mock.Mock()
        self.find_image = mock.Mock()
        self.get_image = mock.Mock()
        self.download_image = mock.Mock()
        self.reactivate_image = mock.Mock()
        self.deactivate_image = mock.Mock()
        self.stage_image = mock.Mock()
        self.import_image = mock.Mock()

        self.members = mock.Mock()
        self.add_member = mock.Mock()
        self.remove_member = mock.Mock()
        self.update_member = mock.Mock()

        self.remove_tag = mock.Mock()
        self.metadef_namespaces = mock.Mock()

        self.tasks = mock.Mock()
        self.tasks.resource_class = fakes.FakeResource(None, {})
        self.get_task = mock.Mock()

        self.get_import_info = mock.Mock()

        self.auth_token = kwargs['token']
        self.management_url = kwargs['endpoint']
        self.version = 2.0


class TestImagev2(utils.TestCommand):

    def setUp(self):
        super().setUp()

        self.app.client_manager.image = FakeImagev2Client(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )

        self.app.client_manager.identity = identity_fakes.FakeIdentityv3Client(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )


def create_one_image(attrs=None):
    """Create a fake image.

    :param attrs: A dictionary with all attributes of image
    :type attrs: dict
    :return: A fake Image object.
    :rtype: `openstack.image.v2.image.Image`
    """
    attrs = attrs or {}

    # Set default attribute
    image_info = {
        'id': str(uuid.uuid4()),
        'name': 'image-name' + uuid.uuid4().hex,
        'owner_id': 'image-owner' + uuid.uuid4().hex,
        'is_protected': bool(random.choice([0, 1])),
        'visibility': random.choice(['public', 'private']),
        'tags': [uuid.uuid4().hex for r in range(2)],
    }

    # Overwrite default attributes if there are some attributes set
    image_info.update(attrs)

    return image.Image(**image_info)


def create_images(attrs=None, count=2):
    """Create multiple fake images.

    :param attrs: A dictionary with all attributes of image
    :type attrs: dict
    :param count: The number of images to be faked
    :type count: int
    :return: A list of fake Image objects
    :rtype: list
    """
    images = []
    for n in range(0, count):
        images.append(create_one_image(attrs))

    return images


def create_one_image_member(attrs=None):
    """Create a fake image member.

    :param attrs: A dictionary with all attributes of image member
    :type attrs: dict
    :return: A fake Member object.
    :rtype: `openstack.image.v2.member.Member`
    """
    attrs = attrs or {}

    # Set default attribute
    image_member_info = {
        'member_id': 'member-id-' + uuid.uuid4().hex,
        'image_id': 'image-id-' + uuid.uuid4().hex,
        'status': 'pending',
    }

    # Overwrite default attributes if there are some attributes set
    image_member_info.update(attrs)

    return member.Member(**image_member_info)


def create_one_import_info(attrs=None):
    """Create a fake import info.

    :param attrs: A dictionary with all attributes of import info
    :type attrs: dict
    :return: A fake Import object.
    :rtype: `openstack.image.v2.service_info.Import`
    """
    attrs = attrs or {}

    import_info = {
        'import-methods': {
            'description': 'Import methods available.',
            'type': 'array',
            'value': [
                'glance-direct',
                'web-download',
                'glance-download',
                'copy-image',
            ]
        }
    }
    import_info.update(attrs)

    return _service_info.Import(**import_info)


def create_one_task(attrs=None):
    """Create a fake task.

    :param attrs: A dictionary with all attributes of task
    :type attrs: dict
    :return: A fake Task object.
    :rtype: `openstack.image.v2.task.Task`
    """
    attrs = attrs or {}

    # Set default attribute
    task_info = {
        'created_at': '2016-06-29T16:13:07Z',
        'expires_at': '2016-07-01T16:13:07Z',
        'id': str(uuid.uuid4()),
        'input': {
            'image_properties': {
                'container_format': 'ovf',
                'disk_format': 'vhd'
            },
            'import_from': 'https://apps.openstack.org/excellent-image',
            'import_from_format': 'qcow2'
        },
        'message': '',
        'owner': str(uuid.uuid4()),
        'result': {
            'image_id': str(uuid.uuid4()),
        },
        'schema': '/v2/schemas/task',
        'status': random.choice(
            [
                'pending',
                'processing',
                'success',
                'failure',
            ]
        ),
        # though not documented, the API only allows 'import'
        # https://github.com/openstack/glance/blob/24.0.0/glance/api/v2/tasks.py#L186-L190
        'type': 'import',
        'updated_at': '2016-06-29T16:13:07Z',
    }

    # Overwrite default attributes if there are some attributes set
    task_info.update(attrs)

    return task.Task(**task_info)


def create_tasks(attrs=None, count=2):
    """Create multiple fake tasks.

    :param attrs: A dictionary with all attributes of Task
    :type attrs: dict
    :param count: The number of tasks to be faked
    :type count: int
    :return: A list of fake Task objects
    :rtype: list
    """
    tasks = []
    for n in range(0, count):
        tasks.append(create_one_task(attrs))

    return tasks


class FakeMetadefNamespaceClient:

    def __init__(self, **kwargs):
        self.create_metadef_namespace = mock.Mock()
        self.delete_metadef_namespace = mock.Mock()
        self.metadef_namespaces = mock.Mock()
        self.get_metadef_namespace = mock.Mock()
        self.update_metadef_namespace = mock.Mock()

        self.auth_token = kwargs['token']
        self.management_url = kwargs['endpoint']
        self.version = 2.0


class TestMetadefNamespaces(utils.TestCommand):

    def setUp(self):
        super().setUp()

        self.app.client_manager.image = FakeMetadefNamespaceClient(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )

        self.app.client_manager.identity = identity_fakes.FakeIdentityv3Client(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )


def create_one_metadef_namespace(attrs=None):
    """Create a fake MetadefNamespace member.

    :param attrs: A dictionary with all attributes of metadef_namespace member
    :type attrs: dict
    :return: a list of MetadefNamespace objects
    :rtype: list of `metadef_namespace.MetadefNamespace`
    """
    attrs = attrs or {}

    metadef_namespace_list = {
        'created_at': '2022-08-17T11:30:22Z',
        'display_name': 'Flavor Quota',
        'namespace': 'OS::Compute::Quota',
        'owner': 'admin',
        # 'resource_type_associations': ['OS::Nova::Flavor'],
        # The part that receives the list type factor is not implemented.
        'visibility': 'public',
    }

    # Overwrite default attributes if there are some attributes set
    metadef_namespace_list.update(attrs)
    return metadef_namespace.MetadefNamespace(**metadef_namespace_list)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from openstackclient.tests.functional.identity.v2 import common


class EndpointTests(common.IdentityTests):

    def test_endpoint_create(self):
        self._create_dummy_endpoint()

    def test_endpoint_delete(self):
        endpoint_id = self._create_dummy_endpoint(add_clean_up=False)
        raw_output = self.openstack(
            'endpoint delete %s' % endpoint_id)
        self.assertEqual(0, len(raw_output))

    def test_endpoint_multi_delete(self):
        endpoint_id_1 = self._create_dummy_endpoint(add_clean_up=False)
        endpoint_id_2 = self._create_dummy_endpoint(add_clean_up=False)
        raw_output = self.openstack(
            'endpoint delete ' + endpoint_id_1 + ' ' + endpoint_id_2)
        self.assertEqual(0, len(raw_output))

    def test_endpoint_list(self):
        endpoint_id = self._create_dummy_endpoint()
        raw_output = self.openstack('endpoint list')
        self.assertIn(endpoint_id, raw_output)
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, self.ENDPOINT_LIST_HEADERS)

    def test_endpoint_show(self):
        endpoint_id = self._create_dummy_endpoint()
        raw_output = self.openstack('endpoint show %s' % endpoint_id)
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.ENDPOINT_FIELDS)
